#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <stdlib.h>
#include "portbits.h"
#include "global.h"
#include "lcd.h"
#include "uart.h"
//#include <util/delay.h>

//***************************************************************************
//*																			*
//*						LCD Controller fr 320x240 LCD	V. 2.0				*
//*																			*
//*						mit UART , DRAM und Touch untersttzung				*
//*																			*
//*						3 - 6 Fonts und einem Benutzer Font					*
//*																			*
//*						SW und 4GW Mode (Umschaltbar)						*
//*																			*
//*						Fr ATmega 16/164/32/324/644						*
//*																			*
//*						By Alexej O.										*
//*																			*
//***************************************************************************

#define XTAL_CPU			16000000UL
//#define XTAL_CPU			18432000UL
//#define XTAL_CPU			20000000UL

#define UART_BAUD_RATE0		   115200UL 
#define UART_BAUD_RATE1			57600UL 
#define UART_BAUD_RATE2			56000UL 
#define UART_BAUD_RATE3			38400UL 
#define UART_BAUD_RATE4			28800UL
#define UART_BAUD_RATE5			19200UL
#define UART_BAUD_RATE6			14400UL 
#define UART_BAUD_RATE7			 9600UL
//#define UART_BAUD_RATE8		  128000UL

#if ((defined(COMMAND_TIMEOUT))&&(COMMAND_TIMEOUT>0))
 volatile unsigned char ctimeout;
#endif

#define DOTIMEOUT()	ctimeout=COMMAND_TIMEOUT; while (!uart_data()) if (!ctimeout) goto mainloop

register unsigned char R_RAS asm("r3");
register unsigned char R_CAS asm("r4");
register unsigned char R_RAS_W asm("r5");
register unsigned char R_CAS_W asm("r6");

//uint8_t  charpos=1, charline=1;

volatile unsigned char viewstart=0, viewstart2=0, Graymode=1, GrayCnt=3, YCnt=0;
unsigned char fontnr=3, charcol=0b00110000; // textcol=255, backcol=0, 
unsigned char text_x=2,text_y=12,textb_x=2,textb_y=12;

#ifdef LCDTouch
// Touch var
uint16_t eeadXmin EEMEM = 90, eeadXmax EEMEM = 940, eeadYmin EEMEM = 150, eeadYmax EEMEM = 850;
uint16_t adXmin, adXmax, adYmin, adYmax;

volatile uint8_t tprogpos=0, tdatacount=0, tpossent=0, tmode=0, Touchcount=4, notouchcount=0;

uint16_t touchad_x_min, adX, txsent, ad_X;
uint16_t touchad_x[9];	//, x128, x64, x32, x16, x8, x4, x2, x1;

uint16_t touchad_y_min, adY, tysent, ad_Y;
uint16_t touchad_y[8];	// y256, y128, y64, y32, y16, y8, y4, y2, y1;
#endif
void init_touch(void);

int main(void) __attribute__((OS_main));
int main(void)
{		
	unsigned char temp, c, ypos=0;
	uint8_t xpos=0, xoffs=0;
	uint16_t adtemp;
	
	PORTD=(1<<LCD_OnOff);
	PORTC=C_norm;
	PORTB=0;
	PORTA=AD_P_noTouch;
	
	DDRA=AD_D_noTouch;
	DDRB=0xFF;
	DDRC=0xFF;
	DDRD=0b11111100;

	// ADC aktivieren
	ADMUX = (1<<REFS0);						// interne Referenzspannung nutzen
	ADCSRA = (1<<ADPS2)|(1<<ADPS1) | (1<<ADPS0) | (1<<ADEN);     // Frequenzvorteiler 128 / ADC aktivieren                 
	ADCSRA |= (1<<ADSC);                  // eine ADC-Wandlung
	while (ADCSRA & (1<<ADSC) ) {}
	adtemp = ADCW;
	
	lcd_init();						// LCD Controller initialisieren

	temp=0;
	c=PINA;
	if(c & (1<<JP_0))
		temp+=1;
	if(c & (1<<JP_1))
		temp+=2;
	if(c & (1<<JP_2))
		temp+=4;
	c=temp;
	
#ifdef STARTMON
	lcd_string(36,36,PSTR("UART:       BAUD"),3);
	if (c==0)
		lcd_string(50,36,PSTR("115200"),3);
	else if (c==1)
		lcd_string(48,36,PSTR("57600"),3);
	else if (c==2)
		lcd_string(48,36,PSTR("56000"),3);
	else if (c==3)
		lcd_string(48,36,PSTR("38400"),3);
	else if (c==4)
		lcd_string(48,36,PSTR("28800"),3);
	else if (c==5)
		lcd_string(48,36,PSTR("19200"),3);
	else if (c==6)
		lcd_string(48,36,PSTR("14400"),3);
	else if (c==7)
		lcd_string(48,36,PSTR("9600"),3);
#endif

	if (c==1){
		adtemp=UART_BAUD(UART_BAUD_RATE1, XTAL_CPU);
		temp=UART_BAUD_U2X(UART_BAUD_RATE1, XTAL_CPU);
	}
	else if (c==2){
		adtemp=UART_BAUD(UART_BAUD_RATE2, XTAL_CPU);
		temp=UART_BAUD_U2X(UART_BAUD_RATE2, XTAL_CPU);
	}
	else if (c==3){
		adtemp=UART_BAUD(UART_BAUD_RATE3, XTAL_CPU);
		temp=UART_BAUD_U2X(UART_BAUD_RATE3, XTAL_CPU);
	}
	else if (c==4){
		adtemp=UART_BAUD(UART_BAUD_RATE4, XTAL_CPU);
		temp=UART_BAUD_U2X(UART_BAUD_RATE4, XTAL_CPU);
	}
	else if (c==5){
		adtemp=UART_BAUD(UART_BAUD_RATE5, XTAL_CPU);
		temp=UART_BAUD_U2X(UART_BAUD_RATE5, XTAL_CPU);
	}
	else if (c==6){
		adtemp=UART_BAUD(UART_BAUD_RATE6, XTAL_CPU);
		temp=UART_BAUD_U2X(UART_BAUD_RATE6, XTAL_CPU);
	}
	else if(c==7){
		adtemp=UART_BAUD(UART_BAUD_RATE7, XTAL_CPU);
		temp=UART_BAUD_U2X(UART_BAUD_RATE7, XTAL_CPU);
	}
	else {
		adtemp=UART_BAUD(UART_BAUD_RATE0, XTAL_CPU);
		temp=UART_BAUD_U2X(UART_BAUD_RATE0, XTAL_CPU);
	}

	uart_init(adtemp,temp);

	// TX aktivieren
#ifdef megaXX4
	UCSR0C =(1<<UCSZ01)|(1<<UCSZ00);
	UCSR0B |= (1<<TXEN0);
#else
	UCSRC =(1<<URSEL)|(1<<UCSZ1)|(1<<UCSZ0);
	UCSRB |= (1<<TXEN);
#endif

	sei();
	
#ifdef Touchtest
	init_touch();
	Touch_test();
	for(;;){
		temp=0;
	}
#endif

#ifdef Grafiktest
	uint16_t cx=100;
	uint8_t	cy=100, cr=15, xdir=0, ydir=0;	//0==+ / 1==-
	uint16_t cx2=50;
	uint8_t	cy2=50, xdir2=0, ydir2=0;	//0==+ / 1==-

	for(;;){
		
		if(cx+cr>=319)
			xdir=1;
		else if(cx-cr==0)
			xdir=0;
		if(cy+cr>=239)
			ydir=1;
		else if(cy-cr==0)
			ydir=0;
			
		if(xdir==0)
			cx++;
		else
			cx--;
		if(ydir==0)
			cy++;
		else
			cy--;
			
		if(cx2+cr>=319)
		xdir2=1;
		else if(cx2-cr==0)
		xdir2=0;
		if(cy2+cr>=239)
		ydir2=1;
		else if(cy2-cr==0)
		ydir2=0;
		
		if(xdir2==0)
		cx2++;
		else
		cx2--;
		if(ydir2==0)
		cy2++;
		else
		cy2--;
			
		lcd_circle(cx,cy,cr,255,255);
		lcd_circle(cx2,cy2,cr,0,0);
			
	}
#endif

mainloop:
	for (;;)
	{	
		c=uart_getchar();
		if (c>31)					// Display Text
		{	lcd_writechar(xpos+xoffs,ypos,c,fontnr);
			xpos+=text_x;
			if (xpos>=XSIZE/4)
			{	xpos=0;
				if(ypos>YSIZE-text_y)
					ypos=0;
				else
					ypos+=text_y;					
			}
		}
		else if (c==1)				// Cursor Home
		{	xpos=0;
			ypos=0;
		}
		else if (c==2)				// LCD off
		{	ENABLE_VLCD=0;
		}
		else if (c==3)				// LCD on
		{	ENABLE_VLCD=1;
		}
		else if (c==4)				// LCD LED off
		{	
			TCCR1A &= ~(1<<COM1A1);
			LCD_LED_PIN=0;
		}
		else if (c==5)				// LCD LED on
		{	
			TCCR1A |= (1<<COM1A1);
		}
		else if (c==6)				// LCD Touch on
		{	
			CONVERTW x;
			c=uart_getchar();
#ifdef LCDTouch
			if(c<=1){
				tmode=c;
				if(c==1)
					init_touch();
				tpossent=0;
				tdatacount=0;
				ad_X=0;
				ad_Y=0;
				PORTA=AD_P_Touch_XY;
				DDRA= AD_D_Touch_XY;
				tprogpos=1;
			}
			else if(c==2){
				c=uart_getchar();
				if(c==1 || !(c%2==0))
					Touchcount=c;
				else
					Touchcount=4;	
			}
			else if(c>=3){
				x.b[0]=uart_getchar();
				x.b[1]=uart_getchar();
				if(c==3)
					eeprom_write_word (&eeadXmin, x.w);		// X min
				else if(c==4)
					eeprom_write_word (&eeadXmax, x.w);		// X max
				else if(c==5)
					eeprom_write_word (&eeadYmin, x.w);		// Y min
				else if(c==6)
					eeprom_write_word (&eeadYmax, x.w);		// Y max
			}
#endif
		}
		else if (c==7)				// LCD Touch off
		{	
#ifdef LCDTouch			
			tprogpos=0;
			PORTA=AD_P_noTouch;
			DDRA= AD_D_noTouch;
#endif			
		}
		else if (c==8)				// Backspace
		{	
			if(xpos < text_x )
			{
				xpos=(XSIZE/4-text_x);
				if(ypos < text_y)
					ypos=YSIZE-text_y;
				else	
					ypos-=text_y;				
			}
			else
				xpos-=text_x;
			lcd_writechar(xpos+xoffs,ypos,32,fontnr);
		}
		else if (c==9)				// Set Pixel
		{	CONVERTW x;
			unsigned char y,c;
			x.b[0]=uart_getchar();
			x.b[1]=uart_getchar();
			y=uart_getchar();
			c=uart_getchar();
			lcd_setpixel(x.w,y,c);
		}
		else if (c==10)        		// Line Feed
		{	
			if(ypos>= YSIZE-text_y)
				ypos=0;
			else
				ypos+=text_y;	
		}
		else if (c==11)      	 	// Delete in place
		{	lcd_writechar(xpos+xoffs,ypos,32,fontnr);
		}
		else if (c==12)       		// Form Feed
		{	xpos=0;
			ypos=0;
			lcd_clear();
		}
		else if (c==13)				// Carriage Return
		{	xpos=0;
		}
		else if (c==14)				// Set Brightness
		{	
			OCR1A=uart_getchar();
		}
		else if (c==15)				// Set Contrast
		{	OCR1B=uart_getchar();
			//if (c>100)
			//	c=100;
			//OCR1B=c*2+c/2;
		}
		else if (c==16)				// Load BMP
		{	unsigned char x1,y,xs,x,c,y1,ys,mode;
			if (uart_getchar()==0xAA)
			{	x1=uart_getchar();
    			y1=uart_getchar();
    			xs=uart_getchar();
    			ys=uart_getchar();
				mode=uart_getchar();
    			for (y=0; y<ys; y++)
    			{	for (x=0; x<xs; x++)
    				{
#if ((defined(COMMAND_TIMEOUT))&&(COMMAND_TIMEOUT>0))
						DOTIMEOUT();
#endif
						c=uart_getchar();
						if (mode)
	    					lcd_writebyteg(x+x1,y+y1,c);
						else
	    					lcd_writebyte(x+x1,y+y1,c);
    				}
				}
			}
		}
		else if (c==17)    	    	// Set Cursor Position (Column and Row)
		{	xpos=uart_getchar();
			ypos=uart_getchar();
			if (xpos>=XSIZE/4)
				xpos=0;
			if (ypos>=YSIZE)
				ypos=0;
		}
		else if (c==18)				// Draw Line
		{	CONVERTW x1,x2;
			unsigned char y1,y2;
			x1.b[0]=uart_getchar();
			x1.b[1]=uart_getchar();
			y1=uart_getchar();
			x2.b[0]=uart_getchar();
			x2.b[1]=uart_getchar();
			y2=uart_getchar();
			c=uart_getchar();
			lcd_line (x1.w,y1,x2.w,y2,c);
		}
		else if (c==19)				// Draw Circle
		{	
			CONVERTW x1;
			unsigned char y1,c2,r;
			x1.b[0]=uart_getchar();
			x1.b[1]=uart_getchar();
			y1=uart_getchar();
			r=uart_getchar();
			c=uart_getchar();
			c2=uart_getchar();
			lcd_circle (x1.w,y1,r,c,c2);
		}
		else if (c==20)				// Invert Block
		{
			CONVERTW x1,x2;
			unsigned char y1,y2;
			x1.b[0]=uart_getchar();
			x1.b[1]=uart_getchar();
			y1=uart_getchar();
			x2.b[0]=uart_getchar();
			x2.b[1]=uart_getchar();
			y2=uart_getchar();
			lcd_invblock (x1.w,y1,x2.w,y2);
		}		
		else if (c==21)				// Textcolor
		{	
			
			c=uart_getchar();  //Textcol
			temp=uart_getchar();		//Backcol
			charcol=(temp&0b11000000) | ((c&0b11000000)>>2);
		}
		else if (c==22)				// Enable Scroll
		{	//uart_getchar();
			//uart_getchar();
			//uart_getchar();
		}
		else if (c==23)				// Font
		{	
			fontnr=uart_getchar();
			
			if(fontnr == 0){	// User Font
				text_x=textb_x;
				text_y=textb_y;
			}
			else if(fontnr == 1){		// Font 4x6
				text_x=1;
				text_y=6;
			}			
			else if(fontnr == 2){	// Font 8x8
				text_x=2;
				text_y=8;
			}
#if defined(__AVR_ATmega32__) || defined(__AVR_ATmega324P__) || defined(__AVR_ATmega644P__)			
			else if(fontnr == 4){	// Font 12x16
				text_x=3;
				text_y=16;
			}			
			else if(fontnr == 5){	// Font 18x26
				text_x=4;
				text_y=26;
			}
#endif			
#if defined(__AVR_ATmega644P__)			
			else if(fontnr == 6){	// Font 24x40
				text_x=6;
				text_y=40;
			}
#endif						
			else {					// Font 8x12 (default)
				fontnr=3;
				text_x=2;
				text_y=12;
			}
		}
		else if (c==24)				// Farbmodus
		{	c=uart_getchar();
			if(c==0){			// SW mode
				Graymode = 0;
				//viewstart2 = viewstart;
				//lcd_clear();
			}
			else{				// 4GW mode
				Graymode = 1;
				if(viewstart>47)
					 viewstart=0;
				GrayCnt = 1;
				//lcd_clear();
			}
			
		}
		else if (c==25)				// Zeichen in Benutzerzeichensatz schreiben
		{	
			uint8_t y,c,m;
			y=uart_getchar();
			
			m=textb_x*textb_y;
			for (uint8_t i=0; i<m;)
			{
				c=uart_getchar();
				ram_writenibbel(i++,y,c);
				asm volatile ("swap %0" : "=r" (c) : "0" (c));	
				ram_writenibbel(i++,y,c);
			}
		}
		else if (c==26)				// Gre des Benutzerzeichensatz setzen
		{	
			c=uart_getchar();
			textb_x=c/4;
			if(textb_x==0 || textb_x > 10)
				textb_x = 2;
				
			textb_y=uart_getchar();
			if(textb_y==0 || textb_y > 42)
				textb_y = 12;
				
			if(fontnr==0){
				text_x=textb_x;
				text_y=textb_y;
			}	
		}
		else if (c==27)				// Set X Text Offset
		{	
			xoffs=uart_getchar();
			if (Graymode)
				temp=47;
			else
				temp=175; 	
			if (xoffs>temp)
				xoffs=temp;
		}
		else if (c==28)				// Set X Startpos
		{	
			viewstart=uart_getchar();
		}
		else if (c==29)				// Draw Block
		{	CONVERTW x1,x2;
			unsigned char y1,y2,c2;
			x1.b[0]=uart_getchar();
			x1.b[1]=uart_getchar();
			y1=uart_getchar();
			x2.b[0]=uart_getchar();
			x2.b[1]=uart_getchar();
			y2=uart_getchar();
			c=uart_getchar();
			c2=uart_getchar();
			lcd_block (x1.w,y1,x2.w,y2,c,c2);
		}
		else if (c==30)				// Send directly
		{	c=uart_getchar();
			if (c==0)
				lcd_writechar(xpos+xoffs,ypos,uart_getchar(),0);
			else
				lcd_writechar(xpos+xoffs,ypos,uart_getchar(),fontnr);
   			xpos+=text_x;
   			if (xpos>=XSIZE/4)
   			{	xpos=0;
   				if(ypos>YSIZE-text_y)
   					ypos=0;
   				else
   					ypos+=text_y;
			}
		}
		else if (c==31)				// Zeichen an pos.
		{			
			unsigned char n,x,y,c;
			n=uart_getchar();
			x=uart_getchar();
			y=uart_getchar();
			c=uart_getchar();
			if(c>fontcount)
				c=3;
			lcd_writechar(x,y,c,n);
		}
	}
}
#ifdef LCDTouch
void init_touch(void){
	uint16_t temp;
	
	adXmin = eeprom_read_word (&eeadXmin);
	adXmax = eeprom_read_word (&eeadXmax);
	adYmin = eeprom_read_word (&eeadYmin);
	adYmax = eeprom_read_word (&eeadYmax);
	
	touchad_y_min = adYmin*Touchcount;
	temp= (adYmax-adYmin);
	temp= (temp+(temp/15))*Touchcount/2;
	
	for (uint8_t i=0;i<8;i++){
		touchad_y[i]= temp;
		temp= temp/2;
	}

	touchad_x_min = adXmin*Touchcount;
	temp= (adXmax-adXmin);
	temp= (temp+(temp/2)+(temp/10))*Touchcount/2;
	
	for (uint8_t i=0;i<9;i++){
		touchad_x[i]= temp;
		temp= temp/2;
	}	

}
#endif
#ifdef Touchtest
void Touch_test(void){
	lcd_clear();
	lcd_string(6,80,PSTR("Touch Test"),3);
	lcd_string(6,100,PSTR("AD Wert X: "),3);
	lcd_string(6,120,PSTR("AD Wert Y: "),3);
	lcd_string(6,140,PSTR("PX Wert X: "),3);
	lcd_string(6,160,PSTR("PX Wert Y: "),3);
	lcd_string(6,180,PSTR("AD Mode P: "),3);

	tmode=1;
	init_touch();
	tpossent=0;
	tdatacount=0;
	ad_X=0;
	ad_Y=0;
	PORTA=AD_P_Touch_XY;
	DDRA= AD_D_Touch_XY;
	tprogpos=1;
	for(;;){
		lcd_writeint(34,180,tprogpos);
		if(tprogpos == 0){
			
			lcd_writeint(34,100,ad_X);
			lcd_writeint(34,120,ad_Y);
			lcd_writeint(34,140,txsent);
			lcd_writeint(34,160,tysent);
			
			// Daten auf 0 zurcksetzen
			tdatacount =0;
			ad_X=0;
			ad_Y=0;
			//txsent=0;
			//tysent=0;
			tprogpos=1;
			
		}
	}
}

void lcd_writeint(uint8_t x, uint8_t y, uint16_t temp){	
	
	lcd_writechar(x,y,(temp/10000)+0x30,3);
	temp=temp-(temp/10000)*10000;
	lcd_writechar(x+2,y,(temp/1000)+0x30,3);
	temp=temp-(temp/1000)*1000;
	lcd_writechar(x+4,y,(temp/100)+0x30,3);
	temp=temp-(temp/100)*100;
	lcd_writechar(x+6,y,(temp/10)+0x30,3);
	lcd_writechar(x+8,y,(temp%10)+0x30,3);	
	
}
#endif